#include "bool.h"
#include <ctype.h>
#include "entry.h"
#include "intlist.h"
#include <math.h>
#include "sequence.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#define MAX_ALIGNMENT_WIDTH 3000
#define MAXSIGNIF 40
#define MAXPOSITIVES 10000
#define MAXSUBST 100
#define LEFTMARGIN 15
#define RIGHTMARGIN 15
#define HIT_SEGMENT_LENGTH MAX_ALIGNMENT_WIDTH

#define EVALMIN -2

/* the following two arrays will be used for evalue calculations */
int     *block_count_g; /*total number of blocks that can be fit into a sequence of length i */
double  *block_width_g; /*total width of the blocks that can be fit into the sequence */
/* the longest block -- also for evalue calculation */
int     maxblocksize_g;


static int Num_positives;
static Entry_T Entries[MAXPOSITIVES];


static void
initialize_map (int *map) {
  int i;

  for (i = 0; i < 129; ++i) {
    map[i] = 20;
  }
  map['A'] = 0;  map['B'] = 1;  map['C'] = 2;  map['D'] = 3;
  map['E'] = 4;  map['F'] = 5;  map['G'] = 6;  map['H'] = 7;
  map['I'] = 8;  map['J'] = 20; map['K'] = 9;  map['L'] = 10;
  map['M'] = 11;  map['N'] = 12;  map['O'] = 20; map['P'] = 13;
  map['Q'] = 14;  map['R'] = 15;  map['S'] = 16;  map['T'] = 17;
  map['U'] = 20; map['V'] = 18;  map['W'] = 19;  map['X'] = 20;
  map['Y'] = 21;  map['Z'] = 22;  
  
  map['a'] = 0;  map['b'] = 1;  map['c'] = 2;  map['d'] = 3;
  map['e'] = 4;  map['f'] = 5;  map['g'] = 6;  map['h'] = 7;
  map['i'] = 8;  map['j'] = 20; map['k'] = 9;  map['l'] = 10;
  map['m'] = 11;  map['n'] = 12;  map['o'] = 13; map['p'] = 13; 
  map['q'] = 14;  map['r'] = 15;  map['s'] = 16;  map['t'] = 17;
  map['u'] = 20; map['v'] = 18;  map['w'] = 19;  map['x'] = 20;
  map['y'] = 21;  map['z'] = 22;

  /* map['*'] = 23;  map['-'] = 24; */
}


static int *
map_sequence (char *sequence_char, int length) {
  int *sequence_int;
  int map[129], i;

  sequence_int = (int *) calloc(length,sizeof(int));
  initialize_map(map);
  for (i = 0; i < length; ++i) {
    sequence_int[i] = map[(int)sequence_char[i]];
  }
  return sequence_int;
}


static double
interpolate_score (int score, int *thresholds) {
  double prob0, prob1;
  int sig, sig0, sig1, score0, score1;

  sig = 0;
  while (score <= thresholds[sig]) {
    ++sig;
  }
  if (sig == 0) {
    return pow(10.0,-MAXSIGNIF);
  } else {
    sig1 = MAXSIGNIF - sig;
    sig0 = sig1+1;

    score1 = thresholds[MAXSIGNIF - sig1];
    score0 = thresholds[MAXSIGNIF - sig0];
    prob1 = pow(10.0,-sig1);
    prob0 = pow(10.0,-sig0);
    if (score0 == score1) {
      return prob0;
    } else {
      return ((double) (score - score0)/(double) (score1 - score0)) 
	* (prob1 - prob0) + prob0;
    }
  }
}

static void get_block_numwidth(char *data_dir, char *prefix){
    char *filename, *base;
    FILE *mat_fp;
    int blocknum;
    int query_length, width, block;


    int length, baselength = 0;
    baselength += strlen (data_dir);
    if (data_dir[strlen (data_dir) - 1] != '/') {
        baselength += 1;
    }
    baselength += strlen (prefix);

    base = (char *) calloc (baselength+1, sizeof (char));
    strcpy (base, data_dir);
    if (data_dir[strlen (data_dir) - 1] != '/') {
        strcat (base, "/");
    }
    strcat (base, prefix);
    length = strlen (base) + strlen (".mat");
    filename = (char *) calloc (length+1, sizeof (char));
    strcpy (filename, base);
    strcat (filename, ".mat");

    if (!(mat_fp = fopen (filename, "rb"))) {
        fprintf(stderr, "Prefix is %s.\n", prefix);
        fprintf (stderr, "Can't open data file %s for width determination.\n", filename);
        exit (1);
    }
    free(filename);
    free(base);
    fread(&blocknum, sizeof(int), 1, mat_fp);
    fread (&maxblocksize_g, sizeof (int), 1, mat_fp);
    block_count_g = (int *) calloc (maxblocksize_g + 1, sizeof(int));
    block_width_g = (double *) calloc (maxblocksize_g + 1, sizeof(double));

    /* scan through the matrix file taking counts of matrix widths */
    for (block = 0; block < blocknum; ++block) {
        fread (&width, sizeof (int), 1, mat_fp);
        if(width > maxblocksize_g){
            fprintf(stderr, "The %s file appears to be corrupt -- found block of length %d\n",
            filename, width);
            exit(1);
        }
        for(query_length = width; query_length <= maxblocksize_g; query_length++){
            block_count_g[query_length]++;
            block_width_g[query_length] += (double) width;
        }
        fseek(mat_fp, sizeof(float), SEEK_CUR); /* skip info */
        	/* fread (&info, sizeof (float), 1, mat_fp); */
        fseek(mat_fp, sizeof(long), SEEK_CUR); /* skip descript_ptr */
        	/* fread (&descript_ptr, sizeof (long), 1, mat_fp);*/
        fseek(mat_fp,  sizeof (int), SEEK_CUR); /* skip descript_length */
        	/* fread (&descript_length, sizeof (int), 1, mat_fp); */

        fseek(mat_fp,  sizeof (short) * width * 24, SEEK_CUR); /* skip perm and the matrix */
        	/* fread (&perm[0], sizeof (short), width, mat_fp); */
        	/* fread (&matrix[0], sizeof (short), width * 23, mat_fp); */
    }
    if(block_count_g[maxblocksize_g] != blocknum){
        fprintf(stderr, "Block counts don't match: %d vs %d\n", block_count_g[maxblocksize_g], blocknum);
        exit(1);
    }

    fclose (mat_fp);
}




static void
open_data_files (FILE **matrices_file, FILE **thresholds_file,
		 FILE **descripts_file, char *data_dir, char *prefix) {
  char *filename, *base;
  int length, baselength = 0;
  
  baselength += strlen(data_dir);
  if (data_dir[strlen(data_dir)-1] != '/') {
    baselength += 1;
  }
  baselength += strlen(prefix);
  base = (char *) calloc(baselength+1,sizeof(char));
  strcpy(base,data_dir);
  if (data_dir[strlen(data_dir)-1] != '/') {
    strcat(base,"/");
  }
  strcat(base,prefix);

  length = strlen(base)+strlen(".mat");
  filename = (char *) calloc(length+1,sizeof(char));
  strcpy(filename,base);
  strcat(filename,".mat");
  if (!(*matrices_file = fopen(filename,"rb"))) {
    fprintf(stderr,"Can't open data file %s\n",filename);
    exit(1);
  }    
  free(filename);

  length = strlen(base)+strlen(".thr");
  filename = (char *) calloc(length+1,sizeof(char));
  strcpy(filename,base);
  strcat(filename,".thr");
  if (!(*thresholds_file = fopen(filename,"rb"))) {
    fprintf(stderr,"Can't open data file %s\n",filename);
    exit(1);
  }    
  free(filename);

  length = strlen(base)+strlen(".des");
  filename = (char *) calloc(length+1,sizeof(char));
  strcpy(filename,base);
  strcat(filename,".des");
  if (!(*descripts_file = fopen(filename,"rb"))) {
    fprintf(stderr,"Can't open data file %s\n",filename);
    exit(1);
  }    
  free(filename);
  free(base);
}

static void
open_data_files_floor (FILE **matrices_file, FILE **thresholds_file,
		       FILE **floors_file, FILE **descripts_file, 
		       char *data_dir, char *prefix) {
  char *filename, *base;
  int length, baselength = 0;
  
  baselength += strlen(data_dir);
  if (data_dir[strlen(data_dir)-1] != '/') {
    baselength += 1;
  }
  baselength += strlen(prefix);
  base = (char *) calloc(baselength+1,sizeof(char));
  strcpy(base,data_dir);
  if (data_dir[strlen(data_dir)-1] != '/') {
    strcat(base,"/");
  }
  strcat(base,prefix);

  length = strlen(base)+strlen(".mat");
  filename = (char *) calloc(length+1,sizeof(char));
  strcpy(filename,base);
  strcat(filename,".mat");
  if (!(*matrices_file = fopen(filename,"rb"))) {
    fprintf(stderr,"Can't open data file %s\n",filename);
    exit(1);
  }    
  free(filename);

  length = strlen(base)+strlen(".thr");
  filename = (char *) calloc(length+1,sizeof(char));
  strcpy(filename,base);
  strcat(filename,".thr");
  if (!(*thresholds_file = fopen(filename,"rb"))) {
    fprintf(stderr,"Can't open data file %s\n",filename);
    exit(1);
  }    
  free(filename);

  length = strlen(base)+strlen(".flr");
  filename = (char *) calloc(length+1,sizeof(char));
  strcpy(filename,base);
  strcat(filename,".flr");
  if (!(*floors_file = fopen(filename,"rb"))) {
    fprintf(stderr,"Can't open data file %s\n",filename);
    exit(1);
  }    
  free(filename);

  length = strlen(base)+strlen(".des");
  filename = (char *) calloc(length+1,sizeof(char));
  strcpy(filename,base);
  strcat(filename,".des");
  if (!(*descripts_file = fopen(filename,"rb"))) {
    fprintf(stderr,"Can't open data file %s\n",filename);
    exit(1);
  }    
  free(filename);
  free(base);
}

static void
hit_copy (char *segment, char *sequence_char, int start, int width) {
  int l, p;

  for (l = start, p = 0; l < start + width; ++l) {
      segment[p++] = sequence_char[l];
  }
  segment[p] = '\0';
}


static void
score_sequence_ordered (int *sequence_int, char *sequence_char, int length,
			int cutoff_sig, double comparison_no, char *prefix, 
			float threshold_info, int *normalP) {
  FILE *mat_fp, *thresh_fp, *floor_fp, *descripts_fp;
  int i, j, k, block, width, score, total, maxblocksize;
  char hit[HIT_SEGMENT_LENGTH];
  double prob, eval;
  short *matrix, *perm;
  int sig_index, descript_length, maxdescriptlength;
  char *description;
  long descript_ptr, last_descript_ptr = 0L;
  int thresholds[MAXSIGNIF], thresh, *floors;
  float info;
  *normalP = 1;


  sig_index = MAXSIGNIF - cutoff_sig;
  open_data_files_floor(&mat_fp,&thresh_fp,&floor_fp,&descripts_fp,
			DATA_DIR,prefix);
  fread(&total,sizeof(int),1,mat_fp);
  fread(&maxblocksize,sizeof(int),1,mat_fp);
  fread(&maxdescriptlength,sizeof(int),1,descripts_fp);

  description = (char *) calloc(maxdescriptlength+1,sizeof(char));
  matrix = (short *) calloc(maxblocksize*23,sizeof(short));
  floors = (int *) calloc(maxblocksize,sizeof(int));
  perm = (short *) calloc(maxblocksize,sizeof(short));

  Num_positives = 0;
  for (block = 0; block < total; ++block) {
    fread(&width,sizeof(int),1,mat_fp);
    fread(&info,sizeof(float),1,mat_fp);
    fread(&descript_ptr,sizeof(long),1,mat_fp);
    fread(&descript_length,sizeof(int),1,mat_fp);

    fread(&perm[0],sizeof(short),width,mat_fp);
    fread(&matrix[0],sizeof(short),width*23,mat_fp);
    fread(&floors[0],sizeof(int),width,floor_fp);
    fread(&thresholds,sizeof(int),MAXSIGNIF,thresh_fp);

    /* j is position along sequence */
    /* k is position along scoring matrix */
    /* i is a matrix pointer */
    if (info >= threshold_info) {
      thresh = thresholds[sig_index];
      for (j = 0; j <= length - width; ++j) {
	score = -thresh;
	for (k = 0, i = 0; k < width; ++k, i += 23) {
	  score += matrix[i + sequence_int[j + perm[k]]];
	  if (score < floors[k]) {
	    goto too_low;
	  }
	}
	score += thresh;
	if (Num_positives >= MAXPOSITIVES - 1) {
	  /*
	  printf("Error! Too many positives: Limit is %d\n",
		 MAXPOSITIVES);
	  */
	  Num_positives--;
	  *normalP = 0;
	  free(perm);
	  free(floors);
	  free(matrix);
	  free(description);
	  fclose(descripts_fp);
	  fclose(floor_fp);
	  fclose(thresh_fp);
	  fclose(mat_fp);
	  return;
	  /* exit(1); */ 
	} else {
	  hit_copy(hit,sequence_char,j,width);
	  prob = interpolate_score(score,thresholds);
	  eval = prob * comparison_no;
	  fseek(descripts_fp,descript_ptr-last_descript_ptr,SEEK_CUR);
	  fread(&description[0],sizeof(char),descript_length,descripts_fp);
	  last_descript_ptr = descript_ptr + (long) descript_length;
	  Entries[Num_positives++] = 
	    Entry_new(description,descript_length,j,width,score,prob,eval, info,hit);
	}
      too_low: ;
      }
    }
  }

  free(perm);
  free(floors);
  free(matrix);
  free(description);
  fclose(descripts_fp);
  fclose(floor_fp);
  fclose(thresh_fp);
  fclose(mat_fp);
}

void
print_entries () {
  int i;
  
  for (i = 0; i < Num_positives; ++i) {
    Entry_print(Entries[i]);
  }
}


static void
parse_arguments (bool *quietp,
		 int *cutoff_sig, int *cutoff_eval, 
		 float *threshold_info,
		 char **prefix, char **seqfile, int argc, char *argv[]) {
  ++argv; --argc;
  while (argc > 0) {
    if ((*argv)[0] == '-') {
      if ((*argv)[1] == 'e') {
	*cutoff_eval = atoi (argv[1]); argc -= 2; ++argv; ++argv;
	if (*cutoff_eval <= EVALMIN) {
	  fprintf (stderr,
		   "Cutoff evalue significance must be greater than %d\n", 
		   EVALMIN);
	  exit (1);
	} else if (*cutoff_eval > MAXSIGNIF) {
	  fprintf (stderr,
		   "Cutoff evalue significance must be considerably \n(about six orders of magnitude) less than %d\n", 
		   MAXSIGNIF);
	  exit(1);
	}
      } else if ((*argv)[1] == 's') {
	*cutoff_sig = atoi(argv[1]); argc -= 2; ++argv; ++argv;
	if (*cutoff_sig <= 0) {
	  fprintf(stderr,"Cutoff significance must be greater than 0\n");
	  exit(1);
	} else if (*cutoff_sig > MAXSIGNIF) {
	  fprintf(stderr,"Cutoff significance must be %d or less\n",MAXSIGNIF);
	  exit(1);
	}
      } else if ((*argv)[1] == 'Q') {
	*quietp = TRUE; argc -= 1; ++argv;
      } else if ((*argv)[1] == 'i') {
	*threshold_info = atof(argv[1]); argc -= 2; ++argv; ++argv;
      } else if ((*argv)[1] == 'f') {
	++argv; *prefix = *argv; argc -= 2; ++argv;
      } else {
	printf("I don't understand.\n");
	exit(1);
      }
    } else {
      *seqfile = *argv; argc -= 1; argv += 1;
    }
  }
}

int 
main (int argc, char **argv) {
  bool quietp = FALSE;
  bool use_evaluep = FALSE;
  int cutoff_sig = DEFAULT_SEARCH_THRESHOLD, length;
  float threshold_info = 0.0;
  int cutoff_eval = EVALMIN;
  char *seqfile = NULL;
  int c;
  FILE *fp;
  int *sequence_int, i;
  char *sequence_char, *name, *prefix = "blocksplus";
  int blocknum;
  double blockwidth = 1.0;
  double comparison_no = 0.0;
  int normalTerm = 1;

  parse_arguments(&quietp, &cutoff_sig,&cutoff_eval,
  		  &threshold_info,&prefix,&seqfile,
		  argc,argv);
  if (seqfile == NULL) {
    fp = stdin;
  } else {
    if (!(fp = fopen(seqfile,"r"))) {
      fprintf(stderr,"Can't open sequence file %s\n",seqfile);
      exit(1);
    }
  }

  /* we need to determine whether to use p-value or e-value for the cutoff */
  if(cutoff_eval > EVALMIN){
    /* if an evalue cutoff is entered, it will be used regardless of the 
       pvalue*/
    use_evaluep = TRUE;
  }


  /* need this call to convert between probabilities and evalues */
  get_block_numwidth(DATA_DIR, prefix);

  while (!feof(fp) && !isspace(c = fgetc(fp))) {
    ungetc(c,fp);
    if ((sequence_char = Sequence_read(fp,&name,&length))) {
	sequence_int = map_sequence(sequence_char,length);
	if(length < maxblocksize_g){
            /* fprintf(stderr, "The length (%d) is smaller.\n", length); */
            blockwidth = block_width_g[length];
            blocknum = block_count_g[length];
        } else{
            /* fprintf(stderr, "The length (%d) is longer.\n", length); */
            blocknum = block_count_g[maxblocksize_g];
            blockwidth = block_width_g[maxblocksize_g];
        }
        comparison_no = (double) blocknum * ((double)length + 1.0) - blockwidth;
        /* fprintf(stderr, "Length: %d, Number of Comparisons: %f\n", length, comparison_no);    */
	
	if(use_evaluep){
	    cutoff_sig = (int)log10(comparison_no) +  cutoff_eval;
	    if(cutoff_sig > MAXSIGNIF){
		cutoff_sig = MAXSIGNIF;
		/*
		fprintf (stderr, "When calculating probability cutoffs from\n");
		fprintf (stderr, "evalue cutoffs, a probability cutoff exceeded\n");
		fprintf (stderr, "%d \n", MAXSIGNIF);
		exit(1);
		*/
	   }
	}
	score_sequence_ordered( sequence_int,sequence_char,length,cutoff_sig, 
				comparison_no, prefix,threshold_info, &normalTerm);

	free(sequence_int);
	free(sequence_char);

	qsort(Entries,Num_positives,sizeof(Entry_T),Entry_cmp);

	if (!quietp) {
	  if (name != NULL) {
	    printf(">%s\n",name);
		if(!normalTerm){
				printf("*******Too many hits: Did not examine all of them.\n ");
		}
	    free(name);
	  }
	  print_entries();
	}

	for (i = 0; i < Num_positives; ++i) {
	  Entry_free(&(Entries[i]));
	}
    }
  }

  if (seqfile != NULL) {
    fclose(fp);
  }
  return 0;
}

