#include "pattern.h"
#include <stdlib.h>


#define T Pattern_T
struct T {
  int maxorder;
  int order;
  int index;
  int *contents;
};

T
Pattern_new (int index, int order, int maxorder) {
  T pattern = (T) malloc(sizeof(*pattern));
  int i, newindex = 0;

  if (order < 0) {
    order = 0;
  } else if (order > maxorder) {
    order = maxorder;
  }
  pattern->maxorder = maxorder;
  pattern->order = order;
  pattern->contents = (int *) calloc(order,sizeof(int));
  for (i = 0; i < order; ++i) {
    pattern->contents[i] = index % NUM_BASES;
    index = index - NUM_BASES*pattern->contents[i];
  }
  for (i = pattern->order-1; i >= 0; --i) {
    newindex = NUM_BASES*newindex + pattern->contents[i];
  }
  pattern->index = newindex;
  return pattern;
}


void
Pattern_free (T *pattern) {
  free((*pattern)->contents);
  free(*pattern);
}

static int
markov_cum_size (int order, int maxorder) {
  int cum_size = 0, size = 1, i;

  if (order < maxorder) {
    for (i = 0; i < order; ++i) {
      size *= NUM_BASES;
      cum_size += size;
    }
    return cum_size;
  } else {
    for (i = 0; i < maxorder; ++i) {
      size *= NUM_BASES;
      cum_size += size;
    }
    return cum_size;
  }
}

int
Pattern_rel_index (T pattern) {
  return pattern->index;
}

int
Pattern_prob_index (T pattern, int base) {
  return markov_cum_size(pattern->order,pattern->maxorder) + pattern->index + base;
}

